#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


__author__ = 'Denis Mikhalkin'

from dynamofuse.base import BaseRecord
from errno import  ENOENT, EINVAL
from os.path import realpath, join, dirname, basename
from threading import Lock
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError
from time import time
from boto.dynamodb.condition import EQ, GT
from boto.dynamodb.types import Binary
import logging
import cStringIO
from stat import S_IFDIR, S_IFLNK, S_IFREG, S_ISREG, S_ISDIR
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn
import itertools

if not hasattr(__builtins__, 'bytes'):
    bytes = str

class Node(BaseRecord):
    pass


#!/usr/bin/env python

#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import with_statement
from dynamofuse.lock import DynamoLock

__author__ = 'Denis Mikhalkin'

from errno import *
from os.path import realpath
from sys import argv, exit
from threading import Lock
import boto.dynamodb
from posix import F_OK, R_OK
from posix import R_OK, X_OK, W_OK
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError, DynamoDBConditionalCheckFailedError
from boto.exception import BotoServerError, BotoClientError
from boto.exception import DynamoDBResponseError
from stat import *
from boto.dynamodb.types import Binary
from time import time
from boto.dynamodb.condition import EQ, GT
import os
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
import logging
import sys
import cStringIO
import itertools
import traceback

if not hasattr(__builtins__, 'bytes'):
    bytes = str

MAX_RETRIES = 5
DELETED_LINKS="$deleted$"
CONSISTENT_OPER=False

def retry(m):
    def wrappedM(*args):
        retries = 0
        logger = logging.getLogger("dynamo-fuse")
        while retries < MAX_RETRIES:
            try:
                return m(*args)
            except DynamoDBConditionalCheckFailedError, cf:
                logger.debug(cf)
                logger.debug("Retrying " + str(m))
                retries += 1
                if retries >= MAX_RETRIES:
                    raise FuseOSError(EIO)
    return wrappedM

# Note: st_mode, st_gid and st_uid are at inode level
class BaseRecord:
    record = None
    accessor = None
    path = None
    log = logging.getLogger("dynamo-fuse")

    def create(self, accessor, path, attrs):
        self.accessor = accessor
        self.path = path

        name = os.path.basename(path)
        if name == "":
            name = "/"
        l_time = int(time())
        (uid, gid, unused) = fuse_get_context()
        newAttrs = {'name': name, 'path': os.path.dirname(path),
                    'type': self.__class__.__name__,
                    'st_nlink': 1,
                    'st_size': 0, 'st_ctime': l_time,
                    'st_mtime': l_time, 'st_atime': l_time,
                    'st_gid': gid, 'st_uid': uid,
                    'version': 1,
                    'st_ino': attrs['st_ino'] if 'st_ino' in attrs else self.accessor.allocUniqueId()
        }
        for k, v in attrs.items():
            newAttrs[k] = v
        self.log.debug("Create attrs: %s", newAttrs)
        allowOverwrite = "allowOverwrite" in attrs
        if allowOverwrite:
            del attrs["allowOverwrite"]
            item = self.accessor.table.new_item(attrs=newAttrs)
            item.put()
        else:
            item = self.accessor.table.new_item(attrs=newAttrs)
            item.put()

        self.record = item
        logging.getLogger("dynamo-fuse-record").debug("Read record %s, version %d", os.path.join(self.record["path"], self.record["name"]), self.record["version"])
        self.record.save = self.safeSave(self.record, self.record.save)

    def init(self, accessor, path, record):
        self.accessor = accessor
        self.path = path
        self.record = record
        logging.getLogger("dynamo-fuse-record").debug("Read record %s, version %d", os.path.join(record["path"], record["name"]), record["version"])
        self.record.save = self.safeSave(self.record, self.record.save)

    @staticmethod
    def safeSave(record, origSave):
        def safeSaveImpl(**kwargs):
            logging.getLogger("dynamo-fuse-record").debug("Saving record %s, version %d", os.path.join(record["path"], record["name"]), record["version"])
            record.add_attribute("version", 1)
            return origSave(expected_value={"version": record["version"]}, **kwargs)
        return safeSaveImpl

    def getRecord(self):
        return self.record

    def delete(self):
        self.record.delete()

        self.updateDirectoryMCTime(self.path)

    def moveTo(self, newPath, forceUpdate=False):
        self.cloneItem(newPath)

        self.delete()

    def cloneItem(self, path, attrsToPreserve=('type', 'st_nlink', 'st_size', 'st_ino', 'st_dev', 'st_rdev', 'st_mode', 'blockId', 'st_gid', 'st_uid', 'deleted', 'blockId')):
        attrs=dict(self.record)
        del attrs['name']
        del attrs['path']
        if attrsToPreserve:
            toDelete = []
            for attr in attrs.keys():
                if not attr in attrsToPreserve:
                    toDelete.append(attr)
            for attr in toDelete:
                del attrs[attr]
        newItem = self.__class__()
        attrs["allowOverwrite"] = True
        newItem.create(self.accessor, path, attrs)

        self.updateDirectoryMTime(path)
        return newItem

    def getattr(self):
        return self.getRecord()

    def isHidden(self):
        return 'hidden' in self.getRecord()

    @retry
    def chmod(self, mode):
        block = self.getRecord()

        self.permitPrivilegedOrOwner(block)

        block['st_mode'] &= 0770000
        block['st_mode'] |= mode
        self.checkSetGid(block)
        block['st_ctime'] = max(int(time()), block['st_ctime'])
        block.save()

    @retry
    def chown(self, uid, gid):
        block = self.getRecord()

        self.permitPrivilegedOnly(block, uid, gid)
        self.permitOwnerToGroup(block, uid, gid)

        if uid != -1: block['st_uid'] = uid
        if gid != -1: block['st_gid'] = gid
        block['st_ctime'] = max(int(time()), block['st_ctime'])
        block.save()

    def updateCTime(self):
        self.record['st_ctime'] = max(self.record['st_ctime'], int(time()))
        # TODO Concurrency
        self.record.save()

    def updateMTime(self):
        self.record['st_mtime'] = max(self.record['st_mtime'], int(time()))
        # TODO Concurrency
        self.record.save()

    def updateMCTime(self):
        l_time = int(time())
        self.record['st_mtime'] = max(self.record['st_mtime'], l_time)
        self.record['st_ctime'] = max(self.record['st_mtime'], l_time)
        # TODO Concurrency
        self.record.save()

    @retry
    def updateDirectoryMTime(self, filepath):
        dir = self.accessor.getRecordOrThrow(os.path.dirname(filepath))
        dir.updateMTime()

    @retry
    def updateDirectoryMCTime(self, filepath):
        dir = self.accessor.getRecordOrThrow(os.path.dirname(filepath))
        dir.updateMCTime()

    def isFile(self):
        return self.record["type"] == "File"

    def isDirectory(self):
        return self.record["type"] == "Directory"

    @staticmethod
    def isDirectoryItem(item):
        return "type" in item and item['type'] == 'Directory'

    def isLink(self):
        return self.record["type"] == "Symlink"

    def isHardLink(self):
        return self.record["type"] == "Link"

    def isNode(self):
        return self.record["type"] == "Node"

    def getOwner(self):
        return self.getRecord()['st_uid']

    def getParent(self, fs):
        return fs.getRecordOrThrow(os.path.dirname(self.path))

    def takeLock(self):
        return DynamoLock(self.path, self.accessor, self)

    def __getitem__(self, item):
        return self.record[item]

    def __setitem__(self, key, value):
        self.record[key] = value

    def __contains__(self, item):
        return self.record.__contains__(item)

    def save(self):
        self.record.save()

    def access(self, mode):
        if mode == F_OK:
            return 0

        block = self.getRecord()
        st_mode = block['st_mode']
        st_uid = block['st_uid']
        st_gid = block['st_gid']
        return 0 if self.modeAccess(mode, st_mode, st_uid, st_gid) else -1

    def modeToStr(self, mode):
        res = ""
        if mode & S_IRUSR: res += 'r'
        else: res += '-'
        if mode & S_IWUSR: res += 'w'
        else: res += '-'
        if mode & S_IXUSR: res += 'x'
        else: res += '-'

        if mode & S_IRGRP: res += 'r'
        else: res += '-'
        if mode & S_IWGRP: res += 'w'
        else: res += '-'
        if mode & S_IXGRP: res += 'x'
        else: res += '-'

        if mode & S_IROTH: res += 'r'
        else: res += '-'
        if mode & S_IWOTH: res += 'w'
        else: res += '-'
        if mode & S_IXOTH: res += 'x'
        else: res += '-'

        return res

    def modeAccess(self, mode, st_mode, st_uid, st_gid):
        (uid, gid, unused) = fuse_get_context()
#        self.log.debug("modeAccess, node %x(%s) %d %d, input %x %d %d", st_mode, self.modeToStr(st_mode), st_uid, st_gid, mode, uid, gid)
        if mode & R_OK:
            if not (not uid or uid == st_uid and st_mode & S_IRUSR or uid != st_uid and gid == st_gid and st_mode & S_IRGRP or uid != st_uid and gid != st_gid and st_mode & S_IROTH):
                return False

        if mode & W_OK:
            if not (not uid or uid == st_uid and st_mode & S_IWUSR or uid != st_uid and gid == st_gid and st_mode & S_IWGRP or uid != st_uid and gid != st_gid and st_mode & S_IWOTH):
                return False

        if mode & X_OK:
            if not (not uid or uid == st_uid and st_mode & S_IXUSR or uid != st_uid and gid == st_gid and st_mode & S_IXGRP or uid != st_uid and gid != st_gid and st_mode & S_IXOTH):
                return False

        return True

    @retry
    def utimens(self, atime, mtime):
        block = self.getRecord()
        block['st_atime'] = max(int(atime), block['st_atime'])
        block['st_mtime'] = max(int(mtime), block['st_mtime'])
        block.save()

    def permitPrivilegedOnly(self, block, uid, gid):
        (ouid, unused, unused) = fuse_get_context()
        if ouid and not (uid == -1 or uid == block['st_uid']):
            self.log.debug('permitPrivilegedOnly: owner is not privileged %d => %d, owner %d', block['st_uid'], uid, ouid)
            raise FuseOSError(EPERM)

    def permitOwnerToGroup(self, block, uid, gid):
        if block['st_gid'] != gid and gid != -1:
            (ouid, ogid, unused) = fuse_get_context()
            if block['st_uid'] != ouid and ouid or gid != ogid and ogid:
                self.log.debug('permitOwnerToGroup: not allowed %d != %d or %d != %d', block['st_uid'], ouid, gid, ogid)
                raise FuseOSError(EPERM)

    def permitPrivilegedOrOwner(self, block):
        (ouid, unused, unused) = fuse_get_context()
        if block['st_uid'] != ouid and ouid:
            raise FuseOSError(EPERM)

    def checkSetGid(self, block):
        (ouid, ogid, unused) = fuse_get_context()
        if ouid and ogid != block['st_gid']:
            block['st_mode'] &= ~S_ISGID

    def isDeleted(self):
        return False


#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import with_statement
from boto.dynamodb.exceptions import DynamoDBConditionalCheckFailedError

__author__ = 'Denis Mikhalkin'

from errno import EAGAIN, EFAULT
from fuse import FuseOSError
import os
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn
import logging
import cStringIO
import uuid
from time import time, sleep

if not hasattr(__builtins__, 'bytes'):
    bytes = str

MAX_LOCK_RETRIES = 5

class DynamoLock:
    log = logging.getLogger("dynamo-fuse-lock  ")

    def __init__(self, path, accessor, item):
        self.path = path
        self.accessor = accessor
        self.lockId = uuid.uuid4().hex
        self.item = item

    def __enter__(self):
        self.log.debug(" Acquiring exclusive lock on %s", self.path)
        item = self.accessor.newItem(attrs={
            "path": os.path.dirname(self.path),
            "name": os.path.basename(self.path)
        })
        item.put_attribute('lockOwner', self.lockId)
        retries = 0
        while retries < MAX_LOCK_RETRIES:
            try:
                item.save(expected_value={'lockOwner': False})
                self.log.debug(" Got the lock on %s", self.path)
                return
            except DynamoDBConditionalCheckFailedError:
                # Somone acquired the lock before us
                sleep(1)
                retries += 1

        self.log.debug(" CANNOT lock %s", self.path)
        self.__exit__()
        raise FuseOSError(EAGAIN)

    def __exit__(self, type=None, value=None, traceback=None):
        self.log.debug(" Releasing exclusive lock on %s", self.path)
        if "deleted" in self.item:
            self.log.debug(" Not saving lock - item %s was deleted", self.path)
        else:
            item = self.accessor.newItem(attrs={
                "path": os.path.dirname(self.path),
                "name": os.path.basename(self.path),
            })
            item.delete_attribute("lockOwner")
            item.save(expected_value={'lockOwner': self.lockId})


#!/usr/bin/env python

#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import with_statement

__author__ = 'Denis Mikhalkin'

import dynamofuse
from posix import R_OK, X_OK, W_OK
from dynamofuse.records.directory import Directory
from dynamofuse.records.file import File
from dynamofuse.records.node import Node
from dynamofuse.records.symlink import Symlink
from dynamofuse.base import BaseRecord, DELETED_LINKS, CONSISTENT_OPER
from dynamofuse.records.link import Link
from errno import *
from os.path import realpath
from sys import argv, exit
from threading import Lock
import boto.dynamodb
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError, DynamoDBConditionalCheckFailedError
from boto.exception import BotoServerError, BotoClientError
from boto.exception import DynamoDBResponseError
from boto.dynamodb2.table import Table
from stat import *
from boto.dynamodb.types import Binary
from time import time, sleep, clock
from boto.dynamodb.condition import EQ, GT
import os
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
from io import FileIO
import logging
from logging import StreamHandler, FileHandler
import sys
import cStringIO
import itertools
import traceback
from boto.dynamodb2.fields import HashKey, RangeKey, KeysOnlyIndex, AllIndex, IncludeIndex
from boto.dynamodb2.layer1 import DynamoDBConnection
from boto.dynamodb2.table import Table
from boto.dynamodb2.types import NUMBER, STRING

if not hasattr(__builtins__, 'bytes'):
    bytes = str

ALL_ATTRS = None
NAME_MAX = 255 # To match what is expected by Fuse and FSTest
KEY_MAX = 1024
global logStream

class BotoExceptionMixin:
    log = logging.getLogger("dynamo-fuse-oper  ")
    accessLog = logging.getLogger("dynamo-fuse-access")

    def __call__(self, op, path, *args):
        try:
            ret = getattr(self, op)(path, *args)
            self.log.debug("  - %s: %s", op, repr(ret))
            if logStream:
                logStream.flush()
            return ret
        except BotoServerError, e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.log.error("  - %s: %s", op, "".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            raise FuseOSError(EIO)
        except BotoClientError, e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.log.error("  - %s: %s", op, "".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            raise FuseOSError(EIO)
        except DynamoDBResponseError, e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.log.error("  - %s: %s", op, "".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            raise FuseOSError(EIO)
        except FuseOSError, e:
            self.log.error("  - %s: FuseOSError(%s)", op, e.strerror)
            raise e
        except BaseException, e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.log.error("  - %s: %s", op, "".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            raise FuseOSError(EIO)


class DynamoFS(BotoExceptionMixin, Operations):
    BLOCK_SIZE = 32768

    recordTypes = {
        "File": File,
        "Directory": Directory,
        "Symlink": Symlink,
        "Node": Node,
        "Link": Link
    }

    def __init__(self, uri):
        (unused, regionPath) = uri.split(':')
        (region, tableName) = regionPath.split('/')
        self.log = logging.getLogger("dynamo-fuse-oper  ")
        self.tableName = tableName
        self.region = region
        for reg in boto.dynamodb2.regions():
            if reg.name == region:
                self.regionv2 = reg
                break
        self.conn = boto.dynamodb.connect_to_region(region, aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
            aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'])
        connection = DynamoDBConnection(aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
            aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'], region=self.regionv2)
        try:
            self.table = self.conn.get_table(tableName)
            self.tablev2 = Table(tableName, connection=connection)
            self.blockTable = self.conn.get_table(self.tableName + "Blocks")
            self.blockTablev2 = Table(self.tableName + "Blocks", connection=connection)
        except:
            self.createTable()
        self.counter = itertools.count()
        self.__createRoot()
        print "Ready"

    def createTable(self):
        connection = DynamoDBConnection(aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
            aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'], region=self.regionv2)
        self.blockTablev2 = Table.create(self.tableName + "Blocks",
            schema=[
                HashKey('blockId'),
                RangeKey('blockNum', data_type=NUMBER)
            ],
            throughput={'read': 30, 'write': 10},
            connection=connection
        )
        self.tablev2 = Table.create(self.tableName,
            schema=[
                HashKey('path'),
                RangeKey('name')
            ],
            throughput={'read': 30, 'write': 10},
            indexes=[
                KeysOnlyIndex("Links", parts=[
                    HashKey('path'),
                    RangeKey('link')
                ])
            ],
            connection=connection
        )

        description = connection.describe_table(self.tableName)
        iter = 0
        while description["Table"]["TableStatus"] != "ACTIVE":
            print "Waiting for %s to create %d..." % (self.tableName, iter)
            iter += 1
            sleep(1)
            description = connection.describe_table(self.tableName)
        self.table = self.conn.get_table(self.tableName)
        self.blockTable = self.conn.get_table(self.tableName + "Blocks")

    def init(self, conn):
        self.log.debug(" init")

    def __createRoot(self):
        if not self.table.has_item("/", "/"):
            self.mkdir("/", 0755)
        if not self.table.has_item("/", DELETED_LINKS):
            self.mkdir("/" + DELETED_LINKS, 0755)

    def chmod(self, path, mode):
        self.log.debug(" chmod(%s, mode=%d)", path, mode)

        self.checkAccess(os.path.dirname(path), X_OK)

        self.getRecordOrThrow(path).chmod(mode)
        return 0

    def chown(self, path, uid, gid):
        self.log.debug(" chown(%s, uid=%d, gid=%d)", path, uid, gid)

        self.checkAccess(os.path.dirname(path), X_OK)

        self.getRecordOrThrow(path).chown(uid, gid)
        return 0

    def getattr(self, path, fh=None):
        self.log.debug(" getattr(%s)", path)

        self.checkAccess(os.path.dirname(path), X_OK)

        record = self.getRecordOrThrow(path)
        if record.isHidden():
            raise FuseOSError(ENOENT)

        return record.getattr()

    def open(self, path, flags):
        self.log.debug(" open(%s, flags=0x%x)", path, flags)

        access = X_OK
        if flags & os.O_CREAT: access |= W_OK
        self.checkAccess(os.path.dirname(path), access)
        self.checkFileExists(path)

        access = 0
        if flags & (os.O_RDONLY | os.O_RDWR) or flags == 0: access |= R_OK
        if flags & (os.O_WRONLY | os.O_RDWR | os.O_APPEND | os.O_TRUNC): access |= W_OK

        self.checkAccess(path, access)

        return self.allocId()

    def utimens(self, path, times=None):
        self.log.debug(" utimens(%s)", path)
        now = int(time())
        atime, mtime = times if times else (now, now)

        item = self.getRecordOrThrow(path)

        item.utimens(atime, mtime)

    def opendir(self, path):
        self.log.debug(" opendir(%s)", path)

        self.checkAccess(os.path.dirname(path), X_OK)
        self.checkFileExists(path)
        self.checkAccess(path, R_OK | X_OK)

        return self.allocId()

    def readdir(self, path, fh=None):
        self.log.debug(" readdir(%s)", path)
        # Verify the directory exists
        dir = self.getRecordOrThrow(path)

        if dir.access(R_OK | X_OK):
            raise FuseOSError(EACCES)

        yield '.'
        yield '..'
        for v in dir.list(): yield v

    def mkdir(self, path, mode):
        self.log.debug(" mkdir(%s)", path)

        if path != "/":
            self.checkAccess(os.path.dirname(path), R_OK | W_OK | X_OK)

        self.create(path, mode | S_IFDIR)

    def rmdir(self, path):
        self.log.debug(" rmdir(%s)", path)

        item = self.getRecordOrThrow(path)

        if not item.isDirectory():
            raise FuseOSError(EINVAL)

        self.checkAccess(os.path.dirname(path), R_OK | W_OK | X_OK)
        self.checkSticky(path)

        if len(list(item.list())) > 0:
            raise FuseOSError(ENOTEMPTY)

        item.delete()

    def rename(self, old, new):
        self.log.debug(" rename(%s, %s)", old, new)
        if old == new: return
        if old == "/" or new == "/":
            raise FuseOSError(EINVAL)

        self.checkPath(new)
        self.checkAccess(os.path.dirname(old), R_OK | W_OK | X_OK)
        self.checkAccess(os.path.dirname(new), R_OK | W_OK | X_OK)
        self.checkSticky(old, new)

        item = self.getRecordOrThrow(old)
        newItem = self.getRecordOrNone(new)
        if item.isDirectory():
            if not newItem is None:
                if not newItem.isDirectory():
                    raise FuseOSError(EISDIR)
                if not newItem.isEmpty():
                    raise FuseOSError(ENOTEMPTY)

        newDir = self.getItemOrNone(os.path.dirname(new), attrs=['type'])
        if newDir is None or not ('type' in newDir and newDir["type"] == "Directory"):
            raise FuseOSError(ENOENT)

        item.moveTo(new)

    def readlink(self, path):
        self.log.debug(" readlink(%s)", path)

        item = self.getRecordOrThrow(path)

        if not item.isLink():
            raise FuseOSError(EINVAL)

        return item["symlink"]

    def symlink(self, target, source):
        self.log.debug(" symlink(%s, %s)", target, source)

        self.checkAccess(os.path.dirname(target), R_OK | W_OK | X_OK)

        try:
            record = self.createRecord(target, "Symlink", attrs={'symlink': source})
        except DynamoDBConditionalCheckFailedError: # Means the item already exists
            raise FuseOSError(EEXIST)

        record.updateDirectoryMCTime(target)

        return 0

    def create(self, path, mode, fh=None):
        self.log.debug(" create(%s, %d)", path, mode)

        if path != "/":
            self.checkAccess(os.path.dirname(path), R_OK | X_OK | W_OK)

        type = "Node"
        if mode & S_IFDIR == S_IFDIR:
            type = "Directory"
        elif mode & S_IFLNK == S_IFLNK:
            type = "Symlink"
        elif mode & S_IFIFO == S_IFIFO or mode & S_IFBLK == S_IFBLK or mode & S_IFCHR == S_IFCHR or mode & S_IFSOCK == S_IFSOCK:
            type = "Node"
        elif mode & S_IFREG == S_IFREG:
            type = "File"

        attrs = {'st_mode': mode}
        if os.path.basename(path) == DELETED_LINKS:
            attrs['hidden'] = True

        try:
            record = self.createRecord(path, type, attrs=attrs)
        except DynamoDBConditionalCheckFailedError: # Means the item already exists
            raise FuseOSError(EEXIST)

        # Update
        if path != "/":
            record.updateDirectoryMCTime(path)

        return self.allocId()

    def statfs(self, path):
        self.log.debug(" statfs(%s)", path)
        return dict(
            f_bsize=self.BLOCK_SIZE,
            f_frsize=self.BLOCK_SIZE,
            f_blocks=(sys.maxint - 1),
            f_bfree=(sys.maxint - 2),
            f_bavail=(sys.maxint - 2),
            f_files=self.fileCount(),
            f_ffree=sys.maxint - 1,
            f_favail=sys.maxint - 1,
            f_fsid=0,
            f_flag=0,
            f_namemax=NAME_MAX
        )

    def destroy(self, path):
        self.log.debug(" destroy(%s)", path)
        self.table.refresh(wait_for_active=True)

    def truncate(self, path, length, fh=None):
        self.log.debug(" truncate(%s, %d)", path, length)

        item = self.getRecordOrThrow(path)
        if not item.isFile():
            raise FuseOSError(EINVAL)

        self.checkAccess(os.path.dirname(path), X_OK)

        if item.access(W_OK):
            raise FuseOSError(EACCES)

        item.truncate(length)

    def unlink(self, path):
        self.log.debug(" unlink(%s)", path)

        self.checkAccess(os.path.dirname(path), W_OK | X_OK)
        self.checkSticky(path)

        self.getRecordOrThrow(path).delete()

    def write(self, path, data, offset, fh):
        self.log.debug(" write(%s, len=%d, offset=%d)", path, len(data), offset)

        item = self.getRecordOrThrow(path)
        if not item.isFile() and not item.isHardLink():
            raise FuseOSError(EINVAL)

        return item.write(data, offset)

    def read(self, path, size, offset, fh):
        self.log.debug(" read(%s, size=%d, offset=%d)", path, size, offset)

        item = self.getRecordOrThrow(path)
        if not item.isFile() and not item.isHardLink():
            raise FuseOSError(EINVAL)

        return item.read(offset, size)

    def link(self, target, source):
        self.log.debug(" link(%s, %s)", target, source)

        self.checkAccess(source, R_OK)

        item = self.getRecordOrThrow(source)
        if not item.isFile() and not item.isNode() and not item.isHardLink():
            raise FuseOSError(EINVAL)

        if item.isHardLink():
            item = item.getLink()

        self.checkAccess(os.path.dirname(target), R_OK | W_OK | X_OK)
        self.checkAccess(os.path.dirname(source), R_OK | X_OK)

        try:
            record = Link()
            record.createRecord(self, target, {}, item)
        except DynamoDBConditionalCheckFailedError: # Means the item already exists
            raise FuseOSError(EEXIST)

        item.updateDirectoryMCTime(source)
        record.updateDirectoryMCTime(target)

        return 0

    def lock(self, path, fip, cmd, lock):
        self.log.debug(" lock(%s, fip=%x, cmd=%d, lock=(start=%d, len=%d, type=%x))", path, fip, cmd, lock.l_start,
            lock.l_len, lock.l_type)
        return 0
        # Lock is optional if no concurrent access is expected
        # raise FuseOSError(EOPNOTSUPP)

    def bmap(self, path, blocksize, idx):
        self.log.debug(" bmap(%s, blocksize=%d, idx=%d)", path, blocksize, idx)
        raise FuseOSError(EOPNOTSUPP)

    def mknod(self, path, mode, dev):
        self.log.debug(" mknod(%s, mode=%d, dev=%d)", path, mode, dev)

        self.checkAccess(os.path.dirname(path), R_OK | W_OK | X_OK)

        try:
            record = self.createRecord(path, "File", attrs={'st_mode': mode, 'st_rdev': dev})
        except DynamoDBConditionalCheckFailedError: # Means the item already exists
            raise FuseOSError(EEXIST)

        # Update
        if path != "/":
            record.updateDirectoryMCTime(path)

        return 0

    def access(self, path, amode):
        (uid, gid, unused) = fuse_get_context()
        self.accessLog.debug(" access(%s, mode=%d) by (%d, %d)", path, amode, uid, gid)
        item = self.getRecordOrThrow(path)
        return item.access(amode)

        # ============ PRIVATE ====================

    def checkSticky(self, old, new=None):
        (uid, gid, unused) = fuse_get_context()
        oldDir = self.getRecordOrThrow(os.path.dirname(old))

        if oldDir.isSticky():
            oldItem = self.getRecordOrThrow(old)
            if uid != 0 and not (uid == oldDir.getOwner() or uid == oldItem.getOwner()):
                raise FuseOSError(EPERM)

        if new:
            newItem = self.getRecordOrNone(new)
            if newItem:
                parent = newItem.getParent(self)
                if parent.isSticky():
                    if uid:
                        if not (newItem.getOwner() == uid or parent.getOwner() == uid):
                            raise FuseOSError(EPERM)

    def checkAccess(self, path, mode):
        if not self.access(path, mode) == 0:
            raise FuseOSError(EACCES)

    def checkPath(self, path):
        if len(path) > 4096:
            raise FuseOSError(ENAMETOOLONG)

        name = os.path.basename(path)
        if name == "":
            if os.path.dirname(path) != "/":
                raise FuseOSError(EINVAL)
            name = "/"
        if len(name) > min(NAME_MAX, KEY_MAX):
            raise FuseOSError(ENAMETOOLONG)
        if len(os.path.dirname(path)) > KEY_MAX:
            raise FuseOSError(ENAMETOOLONG)

    def absPath(self, file, refDir):
        if file.startsWith('/'):
            return file
        return os.path.join(refDir, file)

    def fileCount(self):
        self.table.refresh()
        return self.table.item_count

    def allocId(self):
        return self.counter.next()

    def checkFileExists(self, filepath):
        self.getItemOrThrow(filepath, attrs=[])

    def newItem(self, attrs):
        return self.table.new_item(attrs=attrs)

    def getItemOrThrow(self, filepath, attrs=None):
        self.checkPath(filepath)
        if attrs is not None:
            if not "name" in attrs: attrs.append("name")
            if not "path" in attrs: attrs.append("path")
        name = os.path.basename(filepath)
        if name == "":
            name = "/"
        try:
            return self.table.get_item(os.path.dirname(filepath), name, attributes_to_get=attrs, consistent_read=CONSISTENT_OPER)
        except DynamoDBKeyNotFoundError:
            raise FuseOSError(ENOENT)

    def getItemOrNone(self, path, attrs=None):
        self.checkPath(path)
        if attrs is not None:
            if not "name" in attrs: attrs.append("name")
            if not "path" in attrs: attrs.append("path")
        name = os.path.basename(path)
        if name == "":
            name = "/"
        try:
            return self.table.get_item(os.path.dirname(path), name, attributes_to_get=attrs, consistent_read=CONSISTENT_OPER)
        except DynamoDBKeyNotFoundError:
            return None

    def getRecordOrThrow(self, filepath, attrs=None, ignoreDeleted=False):
        self.checkPath(filepath)
        if attrs is not None:
            for i in ["name", "path", "type", "version"]:
                if not i in attrs: attrs.append(i)
        name = os.path.basename(filepath)
        if name == "":
            name = "/"
        try:
            res = self.initRecord(filepath,
                self.table.get_item(os.path.dirname(filepath), name, attributes_to_get=attrs, consistent_read=CONSISTENT_OPER))
            if not ignoreDeleted and res.isDeleted():
                raise FuseOSError(ENOENT)
            return res
        except DynamoDBKeyNotFoundError:
            raise FuseOSError(ENOENT)

    def getRecordOrNone(self, path, attrs=None, ignoreDeleted=False):
        self.checkPath(path)
        if attrs is not None:
            for i in ["name", "path", "type", "version"]:
                if not i in attrs: attrs.append(i)
        name = os.path.basename(path)
        if name == "":
            name = "/"
        try:
            res = self.initRecord(path,
                self.table.get_item(os.path.dirname(path), name, attributes_to_get=attrs, consistent_read=CONSISTENT_OPER))
            if not ignoreDeleted and res.isDeleted():
                raise FuseOSError(ENOENT)
            return res
        except DynamoDBKeyNotFoundError:
            return None

    def initRecord(self, path, item):
        record = self.recordTypes[item['type']]()
        record.init(self, path, item)
        return record

    def createRecord(self, path, type, attrs=None):
        record = self.recordTypes[type]()
        record.create(self, path, attrs)
        return record

    def allocUniqueId(self):
        idItem = self.table.new_item(attrs={'name': 'counter', 'path': 'global'})
        idItem.add_attribute("value", 1)
        res = idItem.save(return_values="ALL_NEW")
        return res["Attributes"]["value"]


def cleanup(uri):
    (unused, regionPath) = uri.split(':')
    (region, tableName) = regionPath.split('/')
    conn = boto.dynamodb.connect_to_region(region, aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
        aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'])
    table = conn.get_table(tableName)
    for item in table.scan(attributes_to_get=["name", "path"]):
        if item["path"] == "/" and item["name"] == "/": continue
        if item["path"] == "global" and item["name"] == "counter": continue
        if item["path"] == '/' and item['name'] == DELETED_LINKS: continue
        item.delete()

if __name__ == '__main__':
    if len(argv) != 3:
        print('usage: %s aws:<region>/<dynamo table> <mount point>' % argv[0])
        exit(1)

    logStream = open('/var/log/dynamo-fuse.log', 'w', 0)
    logging.basicConfig(stream=logStream)
    logging.getLogger("dynamo-fuse-oper  ").setLevel(logging.DEBUG)
    logging.getLogger("dynamo-fuse-access").setLevel(logging.INFO)
    logging.getLogger("dynamo-fuse-record").setLevel(logging.INFO)
    logging.getLogger("dynamo-fuse-file  ").setLevel(logging.DEBUG)
    logging.getLogger("fuse.log-mixin").setLevel(logging.INFO)
    logging.getLogger("dynamo-fuse-lock  ").setLevel(logging.DEBUG)
    logging.getLogger("dynamo-fuse-master").setLevel(logging.DEBUG)
    logging.getLogger("dynamo-fuse-block ").setLevel(logging.DEBUG)

    if argv[2] == "cleanup":
        cleanup(argv[1])
    elif argv[2] == "createTable":
        DynamoFS(argv[1]).createTable()
    else:
        fuse = FUSE(DynamoFS(argv[1]), argv[2], foreground=True, nothreads=True, default_permissions=False,
            auto_cache=False,
            noauto_cache=True, kernel_cache=False, direct_io=True, allow_other=True, use_ino=True, attr_timeout=0)





#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = 'Denis Mikhalkin'

from posix import R_OK, X_OK, W_OK
from dynamofuse.records.block import BlockRecord
from dynamofuse.base import BaseRecord, DELETED_LINKS
from errno import  ENOENT, EINVAL, EPERM
import os
from os.path import realpath, join, dirname, basename
from threading import Lock
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError
from time import time
from boto.dynamodb.condition import EQ, GT
from boto.dynamodb.types import Binary
import logging
import cStringIO
from stat import *
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
import itertools
import uuid

if not hasattr(__builtins__, 'bytes'):
    bytes = str

class File(BaseRecord):
    log = logging.getLogger("dynamo-fuse-master")

    def create(self, accessor, path, attrs):
        self.path = path
        self.accessor = accessor

        assert 'st_mode' in attrs

        if not 'st_blksize' in attrs:
            attrs['st_blksize'] = accessor.BLOCK_SIZE

        if not "blockId" in attrs:
            attrs["blockId"] = str(self.accessor.allocUniqueId())
            attrs["st_ino"] = int(attrs["blockId"])

        BaseRecord.create(self, accessor, path, attrs)

    def getFirstBlock(self, getData=False):
        return self.record

    def getBlock(self, blockNum, getData=False, forUpdate=False):
        return BlockRecord(self.accessor, os.path.join(self.record["blockId"], str(blockNum))).read(getData, forUpdate)

    def createBlock(self, blockNum):
        return BlockRecord(self.accessor, os.path.join(self.record["blockId"], str(blockNum))).create(attrs={
            "blockId": self.record["blockId"], "blockNum": blockNum
        })

    def updateCTime(self):
        block = self.getFirstBlock()
        block['st_ctime'] = max(block['st_ctime'], int(time()))
        block.save()

    def getRecord(self):
        return self.record

    def isDeleted(self):
        return 'deleted' in self.record and self.record['deleted']

    def link(self):
        # TODO: Locking? What if the file is deleted?
        block = self.getFirstBlock()
        block.add_attribute("st_nlink", 1)
        block['st_ctime'] = max(block['st_ctime'], int(time()))
        block.save()

    ################# OPERATIONS ##########################

    def getattr(self):
        block = self.getFirstBlock()
        block["st_blocks"] = (block["st_size"] + self.record["st_blksize"] - 1) / self.record["st_blksize"]
        block["st_ino"] = int(self.record["blockId"])
        return block

    def delete(self):
        with self.takeLock():
            self.deleteFile()

    def deleteFile(self, linked=False):
        block = self.getFirstBlock()
        self.log.debug("Delete file, linked=%s, links=%d", linked, block["st_nlink"])
        block.add_attribute("st_nlink", -1)
        block['st_ctime'] = max(block['st_ctime'], int(time()))
        delete = not linked and ('deleted' in block and not block['deleted'] or not 'deleted' in block)
        block['deleted'] = not linked or ('deleted' in block and block['deleted'])
        if block["st_nlink"] == 1:
            self.log.debug("No more links - deleting records")
            items = self.accessor.blockTablev2.query(blockId__eq=self.record["blockId"], attributes=['blockId', 'blockNum'])
            for entry in items:
                entry.delete()

            BaseRecord.delete(self)
        else:
#            resp = block.save(return_values='ALL_NEW')
#            self.log.debug('After deleting: ' + str(resp))
            if delete:
                block['st_nlink'] -= 1
                self.moveTo(os.path.join("/" + DELETED_LINKS, uuid.uuid4().hex), forceUpdate=True)
            else:
                block.save()

    def write(self, data, offset):
        with self.takeLock():
            self._write(data, offset)
            block = self.getFirstBlock()
            block["st_size"] = max(block["st_size"], offset + len(data))
            block['st_ctime'] = max(block['st_ctime'], int(time()))
            block['st_mtime'] = max(block['st_mtime'], int(time()))
            block.save()

            return len(data)

    def _write(self, data, offset):
        startBlock = offset / self.accessor.BLOCK_SIZE
        endBlock = (offset + len(data) - 1) / self.accessor.BLOCK_SIZE
        initialBlockOffset = self.accessor.BLOCK_SIZE - (offset % self.accessor.BLOCK_SIZE)
        blockOffset = 0
        self.log.debug("write start=%d, last=%d, initial offset %d", startBlock, endBlock, initialBlockOffset)
        for blockNum in range(startBlock, endBlock + 1):
            try:
                block = self.getBlock(blockNum, getData=True)
            except FuseOSError, fe:
                if fe.errno == ENOENT:
                    self.log.debug("write block %d is None", blockNum)
                    block = self.createBlock(blockNum)
                else:
                    raise
            dataSlice = data[0:initialBlockOffset] if blockNum == startBlock else\
            data[blockOffset: blockOffset + self.accessor.BLOCK_SIZE]

            self.log.debug("write block %d slice length %d from offset %d", blockNum, len(dataSlice), blockOffset)
            blockOffset += len(dataSlice)

            block.writeData((offset % self.accessor.BLOCK_SIZE) if blockNum == startBlock else 0, dataSlice)
            block.save()

    def read(self, offset, size):
        startBlock = offset / self.accessor.BLOCK_SIZE
        if offset+size > self.record["st_size"]:
            size = self.record["st_size"] - offset
        endBlock = (offset + size - 1) / self.accessor.BLOCK_SIZE
        data = cStringIO.StringIO()
        try:
            self.log.debug("read blocks [%d .. %d]", startBlock, endBlock)
            for block in range(startBlock, endBlock+1):
                item = self.getBlock(block, getData=True)
                if item is None:
                    self.log.debug("read block %d does not exist", block)
                    break
                if not "data" in item:
                    self.log.debug("read block %d has no data", block)
                    break
                itemData = item["data"].value
                writeLen = min(size, self.accessor.BLOCK_SIZE, len(itemData))
                startOffset = (offset % self.accessor.BLOCK_SIZE) if block == startBlock else 0
                self.log.debug("read block %d has %d data, write %d from %d", block, len(itemData), writeLen,
                    startOffset)
                data.write(itemData[startOffset:startOffset + writeLen])
                size -= writeLen

            return data.getvalue()
        finally:
            data.close()

    def moveTo(self, newPath, forceUpdate=False):
        # Files can be hard-linked. When moved, they will update the targets of their hard-links to point to new name (as hard links are actually by name)
        self.cloneItem(newPath)

        # TODO: A problem - in order to perform this query we have to run it against hash_key + "link" which is impossible with current design
        if self.record["st_nlink"] > 1 or forceUpdate:
            self.log.debug("Retargeting links from %s to %s" % (self.path, newPath))
            items = self.accessor.table.scan({"link":EQ(self.path)})
            for link in items:
                link["link"] = newPath
                link.save()

            # This will force delete the block
#            self.record["st_nlink"] = 1

        self.record.delete()
        self.deleted = True

    def truncate(self, length, fh=None):
        with self.takeLock():
            lastBlock = length / self.accessor.BLOCK_SIZE
            l_time = int(time())

            items = self.accessor.blockTablev2.query(blockId__eq=self.record["blockId"], blockNum__gt=lastBlock, attributes=["blockId", "blockNum"])
            for entry in items:
                entry.delete()

            if length:
                try:
                    lastItem = self.getBlock(lastBlock, getData=True)
                    if lastItem is not None and "data" in lastItem:
                        lastItem['data'] = Binary(lastItem['data'].value[0:(length % self.accessor.BLOCK_SIZE)])
                        lastItem.save()
                except FuseOSError, fe:
                    # Block is missing - so nothing to update
                    if fe.errno == ENOENT:
                        pass
                    else:
                        raise fe

            item = self.getFirstBlock()
            item['st_size'] = length
            item['st_ctime'] = max(l_time, item['st_ctime'])
            item['st_mtime'] = max(l_time, item['st_mtime'])
            item.save()



from __future__ import with_statement

__author__ = 'Denis Mikhalkin'

from posix import R_OK, X_OK, W_OK
from errno import *
from os.path import realpath
from sys import argv, exit
from threading import Lock
import boto.dynamodb
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError, DynamoDBConditionalCheckFailedError
from boto.exception import BotoServerError, BotoClientError
from boto.exception import DynamoDBResponseError
from stat import *
from boto.dynamodb.types import Binary
from time import time, sleep
from boto.dynamodb.condition import EQ, GT
import os
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
from io import FileIO
import logging
from logging import StreamHandler, FileHandler
import sys
import cStringIO
import itertools
import traceback
from threading import Thread

if not hasattr(__builtins__, 'bytes'):
    bytes = str

def performTest():
    conn = boto.dynamodb.connect_to_region("ap-southeast-2", aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
        aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'])
    table = conn.get_table("DynamoFS")

    item1 = table.new_item("/", "b", attrs={"flag":"item1", "version":1})
    item1.put()
    item1 = table.get_item("/", "b")
    table.get_item("/", "b").delete()
    item1["flag"] = "updated"
#    item1.save(expected_value={"path":item1["path"], "name":item1["name"]})
    try:
        item1.save(expected_value={"version":item1["version"]})
    except DynamoDBConditionalCheckFailedError:
        item1 = table.get_item("/", "b")

if __name__ == '__main__':
    performTest()


#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = 'Denis Mikhalkin'

from posix import R_OK, X_OK, W_OK
from dynamofuse.records.block import BlockRecord
from dynamofuse.base import BaseRecord
from errno import  ENOENT, EINVAL, EPERM
import os
from os.path import realpath, join, dirname, basename
from threading import Lock
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError
from time import time
from boto.dynamodb.condition import EQ, GT
from boto.dynamodb.types import Binary
import logging
import cStringIO
from stat import *
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn, fuse_get_context
import itertools

if not hasattr(__builtins__, 'bytes'):
    bytes = str

# TODO Rename handling
class Link(BaseRecord):

    def createRecord(self, accessor, path, attrs, link):
        self.link = link
        attrs['link'] = link.path
        BaseRecord.create(self, accessor, path, attrs)
        self.updateLink()

    def getRecord(self):
        return self.link.getRecord()

    def getLink(self):
        return self.link

    def init(self, accessor, path, record):
        self.accessor = accessor
        self.path = path
        self.record = record
        self.readLink()

    def readLink(self):
        self.link = self.accessor.getRecordOrThrow(self.record['link'], attrs=None, ignoreDeleted=True)

    def updateLink(self):
        self.link.link()

    def delete(self):
        self.link.deleteFile(True)
        BaseRecord.delete(self)

    def read(self, offset, size):
        return self.link.read(offset, size)

    def write(self, data, offset):
        return self.link.write(data, offset)


#!/usr/bin/env python

from __future__ import with_statement

from setuptools import setup
from os import chmod
import stat

import subprocess

try:
    from lib2to3 import refactor
    fixers = set(refactor.get_fixers_from_package('lib2to3.fixes'))
except ImportError:
    fixers = set()

with open('README.md') as readme:
    documentation = readme.read()

setup(
    name = 'dynamo-fuse',
    version = '0.1.3',

    description = 'Linux FUSE file system implementation with AWS DynamoDB as the storage',
    long_description = documentation,
    author = 'Denis Mikhalkin',
    author_email = 'denismo@yahoo.com',
    maintainer = 'Denis Mikhalkin',
    maintainer_email = 'denismo@yahoo.com',
    license = 'GNU General Public License, version 3',
    url = 'https://github.com/denismo/dynamo-fuse',
    packages=['dynamofuse', 'dynamofuse.records'],
    requires=['fusepy(>2.0)'],
    data_files=[
        ('/sbin', ['data/mount.fuse.dynamo'])
    ],

    use_2to3 = True,
    # only use the following fixers (everything else is already compatible)
    use_2to3_exclude_fixers = fixers - set([
        'lib2to3.fixes.fix_except',
        'lib2to3.fixes.fix_future',
        'lib2to3.fixes.fix_numliterals',
    ]),

    classifiers = [
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Filesystems',
    ]
)

try:
    chmod('/sbin/mount.fuse.dynamo', stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IXOTH | stat.S_IROTH)
except:
    pass



#    Dynamo-Fuse - POSIX-compliant distributed FUSE file system with AWS DynamoDB as backend
#    Copyright (C) 2013 Denis Mikhalkin
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = 'Denis Mikhalkin'

from errno import EACCES, ENOENT, EINVAL, EEXIST, EOPNOTSUPP, EIO, EAGAIN
from os.path import realpath
from threading import Lock
import boto.dynamodb
from boto.dynamodb.exceptions import DynamoDBKeyNotFoundError
from stat import S_IFDIR, S_IFLNK, S_IFREG
from datetime import datetime, timedelta
from boto.dynamodb.condition import EQ, GT
from boto.dynamodb.types import Binary
import os
from fuse import FUSE, FuseOSError, Operations, LoggingMixIn
import logging
from collections import deque
import sys
import cStringIO

if not hasattr(__builtins__, 'bytes'):
    bytes = str

blockCache = dict()
blockHistory = deque()
blockLog = logging.getLogger("dynamo-fuse-block ")
BLOCK_CACHE_ENABLED=False

class BlockRecord:
    BLOCK_ATTRS = ['version', "blockId", "blockNum"]
    BLOCK_ALL_ATTRS = ["data"] + BLOCK_ATTRS

    log = logging.getLogger("dynamo-fuse-block ")
    item = None
    accessor = None
    path = None

    def __init__(self, accessor, path):
        self.accessor = accessor
        self.path = path

    def read(self, getData=False, forUpdate=False):
        self.item = BlockRecord.getBlockItem(self.accessor, self.path, getData, forUpdate)
        return self

    def create(self, attrs):
        attrs["version"] = 1
        self.item = self.accessor.blockTable.new_item(attrs=attrs)
        self.item.put(expected_value={'blockId':False, 'blockNum':False})
        if BLOCK_CACHE_ENABLED: BlockRecord.cacheItem(self.path, self.item, self.item["version"])
        return self

    def __getitem__(self, key):
        return self.item[key]

    def __setitem__(self, key, value):
        self.item[key] = value
        return value

    def __contains__(self, item):
        return item in self.item

    def save(self):
        self.item.add_attribute("version", 1)
        if BLOCK_CACHE_ENABLED:
            # Delete this properties before they are saved into DynamoDB. They will be restored next time the item is cached
            if "updateTime" in self.item: del self.item["updateTime"]
            if "fullPath" in self.item: del self.item["fullPath"]
        self.item.save()
        if BLOCK_CACHE_ENABLED:
            BlockRecord.cacheItem(self.path, self.item, self.item["version"] + 1)

    def writeData(self, startOffset, dataSlice):
        if "data" in self.item:
            self.log.debug("write block %s has data", self.path)
            itemData = self.item["data"].value
            self.item['data'] = Binary(itemData[0:startOffset] + dataSlice + itemData[startOffset + len(dataSlice):])
        else:
            self.log.debug("write block %s has NO data", self.path)
            self.item['data'] = Binary(dataSlice)

    @staticmethod
    def getBlockItem(accessor, path, getData=False, forUpdate=False):
        try:
            blockItem = accessor.blockTable.get_item(os.path.dirname(path), int(os.path.basename(path)),
                attributes_to_get=(BlockRecord.BLOCK_ALL_ATTRS if getData else BlockRecord.BLOCK_ATTRS))
        except DynamoDBKeyNotFoundError:
            blockItem = None
        cachedBlockItem = BlockRecord.getCachedBlockItem(path, forUpdate) if BLOCK_CACHE_ENABLED else None
        if not blockItem:
            if cachedBlockItem and (getData and "data" in cachedBlockItem or not getData):
                blockLog.debug('Returning cached block item for %s', path)
                return cachedBlockItem
            else:
                blockLog.debug('Unable to find block or cached block for %s', path)
                raise FuseOSError(ENOENT)
        if cachedBlockItem and blockItem["version"] < cachedBlockItem["version"] and (getData and "data" in cachedBlockItem or not getData):
            blockLog.debug('Returning cached block item for %s', path)
            return cachedBlockItem
        else:
            try:
                del blockCache[path]
            except:
                pass
            try:
                blockHistory.remove(blockItem)
            except:
                pass
            return blockItem

    @staticmethod
    def getCachedBlockItem(path, forUpdate=False):
        if forUpdate:
            item = blockCache.pop(path, None)
            # Need to remove the item because the next update will put it in at different offset
            if item:
                try:
                    blockHistory.remove(item)
                except:
                    pass
            return item
        else:
            return blockCache.get(path, None)

    @staticmethod
    def cacheItem(path, item, version):
        blockLog.debug('Caching block for %s', path)
        item["version"] = version
        item["updateTime"] = datetime.now()
        item["fullPath"] = path
        blockCache[path] = item
        blockHistory.append(item)
        BlockRecord.expellExpiredBlocks()

    @staticmethod
    def expellExpiredBlocks():
        now = datetime.now()
        EXPELL_DELTA = timedelta(seconds=2)
        while len(blockHistory) > 0 and (now - blockHistory[0]["updateTime"]) > EXPELL_DELTA:
            block = blockHistory.popleft()
            try:
                del blockCache[block["fullPath"]]
            except:
                pass